-----------------------------------------------------------------------
-- Get local references to certain often-used global objects for better speed
-- With sincere thanks to Torhal and saiket and _NPCScan.Overlay for showing the way
-----------------------------------------------------------------------
local _G = getfenv(0)

local math = _G.math
local table = _G.table

local ipairs = _G.ipairs
local pairs = _G.pairs
local setmetatable = _G.setmetatable
local tonumber = _G.tonumber
local tostring = _G.tostring
local unpack = _G.unpack
local abs, max, floor = math.abs, math.max, math.floor
local twoPi = math.pi * 2;
--local atan2 = math.atan2;

-----------------------------------------------------------------------
-- WoW passes in the addon name and a private table, so grab those
-----------------------------------------------------------------------
local ADDON_NAME, private = ...

--
-- private
-- .canDisplay				Addon is enabled via profile setting, and other requirements are met.  If addon is enabled, but canDisplay is false
--							then the addon will continue to make computations
-- .enableSurveyRecall		When a distance button is pushed, if this is true, remember the map location and direction for the distance indicator
-- .frameCreated			If true, then the Jughead frame has been created, and myButtonFrame is valid
-- .lastDigInSiteSurveyed	If true, then the last dig in a site was just found.  This is used to force redisplay of Archy minimap nodes
-- .archyFound				If true, then archy was found
-- .gm2Found				If true, then GatherMate2 was found
-- .optionsFrame			Holds my configuration frame
-- .haveHiddenArchyIndicators Flag that "remembers" if we've hidden any Archy indicators. Used to save work if un-hiding when none were hidden.
-- .survey_info_this_dig	Table used to hold information about all the unsuccessful survey points this dig, so we can update the distance indicator.

private.canDisplay = true
private.frameCreated = false
private.lastDigInSiteSurveyed = false
private.enableSurveyRecall = false
private.archyFound = false
private.gm2Found = false
private.survey_info_this_dig = {}  -- {x,y,map,level,color,direction}

-- Keep track of the textures that we've created that are currently not in use
local TexturesUnused = CreateFrame( "Frame" );


local Jughead = LibStub("AceAddon-3.0"):NewAddon("Jughead", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0" ) -- , "AceHook-3.0", "AceBucket-3.0"
Jughead.version = _G.GetAddOnMetadata(ADDON_NAME, "Version")

local Astrolabe = _G.DongleStub("Astrolabe-1.0")

-- Allow other people to find me.. if for some reason they want to anyway
_G["Jughead"] = Jughead
local Archy = _G["Archy"]
local GatherMate = LibStub("AceAddon-3.0"):GetAddon("GatherMate2")

local L = LibStub("AceLocale-3.0"):GetLocale("Jughead", true)


private.Options = {
	Version = Jughead.version;
	Modules = {};
	ModulesAlpha = {};
	ModulesExtra = {};
};

private.OptionsDefault = {
	Version = Jughead.version;
	Modules = {};
	ModulesAlpha = {};
	ModulesExtra = {};
	ShowAll = false;
};



local SITES_PER_CONTINENT = 4
local SURVEY_NONE = 0
local SURVEY_GREEN = 1
local SURVEY_YELLOW = 2
local SURVEY_RED = 3
local MIN_DRY_SURVEY_DISTANCE = 4  -- Closer than this, and the artifact would appear

-- Used to set the color of the distance indicator
local DISTANCE_COLOR_TEXCOORDS = {
	[SURVEY_NONE] = {
		0.75, 1.0, 0.0, 1.0
	},
	[SURVEY_GREEN] = {
		0.0, 0.25, 0.0, 1.0
	},
	[SURVEY_YELLOW] = {
		0.25, 0.5, 0.0, 1.0
	},
	[SURVEY_RED] = {
		0.5, 0.75, 0.0, 1.0
	},
}

-- R G B
local OVERLAY_SURVEY_COLORS = {
	[SURVEY_NONE] = {
		[0] = {0, 0, 0}
	},
	[SURVEY_GREEN] = {
		[0] = {0, .3, 0},
		[1] = {0.8, 1.0, 0.9},
		[2] = {0.8, 1.0, 0.8},
		[3] = {0.7, 1.0, 0.7},
		[4] = {0.6, 1.0, 0.6},
		[5] = {0.5, 1.0, 0.5},
		[6] = {0.4, 1.0, 0.4},
		[7] = {0.3, 1.0, 0.3},
		[8] = {0.2, 1.0, 0.2},
		[9] = {0.1, 1.0, 0.1},
	},
	[SURVEY_YELLOW] = {
		[0] = {.35, .35, 0},
		[1] = {1.0, 1.0, 0.8},
		[2] = {1.0, 1.0, 0.7},
		[3] = {1.0, 1.0, 0.6},
		[4] = {1.0, 1.0, 0.5},
		[5] = {1.0, 1.0, 0.4},
		[6] = {1.0, 1.0, 0.3},
		[7] = {1.0, 1.0, 0.2},
		[8] = {1.0, 1.0, 0.1},
		[9] = {1.0, 1.0, 0.0},
	},
	[SURVEY_RED] = {
		[0] = {.35, 0, 0},
		[1] = {1.0, 0.95, 0.9},
		[2] = {1.0, 0.9, 0.9},
		[3] = {1.0, 0.8, 0.8},
		[4] = {1.0, 0.7, 0.7},
		[5] = {1.0, 0.6, 0.6},
		[6] = {1.0, 0.5, 0.5},
		[7] = {1.0, 0.4, 0.4},
		[8] = {1.0, 0.3, 0.3},
		[9] = {1.0, 0.2, 0.2},
	},
}

local PROFILE_DEFAULTS = {
	profile = {
		general = {
			version = 1,
			enabled = true,
			useArchy = true,
			useGM2 = true,
			autoshow = false
		},
		frame = {
			visible = true,
			locked = true,
			--attachToArchy = true,
			anchor = "TOPLEFT",
			relativeTo = "TOPLEFT",
			positionX = 100,
			positionY = 400,
			scale = 1.0,
			alpha = 1
		},
		Options = {}
	},
}


-------------------------------------------------------------------
-- Interface settings
-------------------------------------------------------------------
Jughead.options =
{
	type = "group",
	name = "Jughead",
	args = {
		enableJH = {
			type = "toggle",
			order = 1,
			name = L["ENABLE"],
			desc = L["ENABLE_DESC"],
			get = function(info) return private.db.general.enabled end,
			set = function(info, val)
				private.db.general.enabled = val
				Jughead:ConfigUpdated()
				end,
			width = "double"
		},
		framescale = {
			type = "range",
			order = 2,
			name = L["FRAMESCALE"],
			desc = L["FRAMESCALE_DESC"],
			min = 0.4,
			max = 1.5,
			step = 0.05,
			get = function(info) return private.db.frame.scale end,
			set = function(info, val)
					private.db.frame.scale = val
					Jughead:ConfigUpdated()
				end,
		},
		lockFrame = {
			type = "toggle",
			order = 13,
			name = L["LOCKED"],
			desc = L["LOCKED_DESC"],
			get = function(info) return private.db.frame.locked end,
			set = function(info, val)
					private.db.frame.locked = val

				end,
		},
		autoshow = {
			type = "toggle",
			order = 14,
			name = L["AUTOSHOW"],
			desc = L["Jughead will autoshow on dig."],
			get = function(info) return private.db.general.autoshow end,
			set = function(info, val)
					private.db.general.autoshow = val
				end,
		},
		useArchy = {
			type = "toggle",
			order = 24,
			name = L["USEARCHY"],
			desc = L["USEARCHY_DESC"],
			get = function(info) return private.db.general.useArchy end,
			set = function(info, val)
					private.db.general.useArchy = val
					if not val then
						private.db.frame.attachToArchy = val
					end
					if private.frameCreated then
						Jughead:ConfigUpdated()
					end

				end,
		},
		useGM2 = {
			type = "toggle",
			order = 25,
			name = L["USEGM2"],
			desc = L["USEGM2_DESC"],
			get = function(info) return private.db.general.useGM2 end,
			set = function(info, val)
					private.db.general.useGM2 = val
				end,
		}
	},
}


-----------------------------------------------------------------------
-- "Prototype" some functions so we can use them
-----------------------------------------------------------------------
local dumpMinimapStats
local fetchArchySurveyPOI
local fetchGM2SurveyPOI
--local CacheMapData

-----------------------------------------------------------------------
-- Some variables for keeping track of things
-----------------------------------------------------------------------
local current_continent  -- Where the heck am I anyway?  Updated by timer
local timer_handle -- used to keep track of player location
local race_data = {}  -- used to keep track of how many fragments we have of each race, so we can know when to turn POI frames back on.
local myButtonFrame   -- holds my frame of red/yellow/green/reset buttons
local SURVEY_YARDS = {[SURVEY_GREEN] = 40, [SURVEY_YELLOW] = 80, [SURVEY_RED] = 800}  -- Need a little wiggle room with these.
local SURVEY_ANGLES = {[SURVEY_GREEN] = {left=math.rad(25),right=math.rad(335)}, [SURVEY_YELLOW] = {left=math.rad(30),right=math.rad(330)}, [SURVEY_RED] = {left=math.rad(50),right=math.rad(310)}}
local archDigFrame -- The Archy dig site frame
local nearestDigsiteIndexer  -- The DIG_SITES index for the site that the player is closest to
local nextNearestDigsiteIndexer  -- The DIG_SITES index for the site that the player is closest to
local player_map, player_level, player_x, player_y -- updated by every UpdatePlayerPosition (5 second timer), or when the dig spell is successfully cast
local ZONE_DATA = {}
local MAP_CONTINENTS = {}



----------------------------------------
----------------------------------------
local function DebugPrint(msg)
--	Jughead:Print(msg)
end

private.DebugPrint = DebugPrint

----------------------------------------
----------------------------------------
local function NoProtectedAllowed()
	return (private.in_combat or _G.InCombatLockdown() or (_G.UnitAffectingCombat("player") or _G.UnitAffectingCombat("pet")))
end


-- With deep thanks to saiket and _NPCScan.Overlay

--- Prepares an unused texture on the given frame.
-- @param Layer  Draw layer for texture.
-- @param ...  Color and optional alpha to set texture to.
function private:TextureCreate ( Layer, R, G, B, A )
	local Texture = #TexturesUnused > 0 and TexturesUnused[ #TexturesUnused ];
	if ( Texture ) then
		TexturesUnused[ #TexturesUnused ] = nil;
		Texture:SetParent( self );
		Texture:SetDrawLayer( Layer );
		Texture:ClearAllPoints();
		Texture:Show();
	else
		Texture = self:CreateTexture( nil, Layer );
	end
	Texture:SetVertexColor( R, G, B, A or 1 );
	Texture:SetBlendMode( "BLEND" );

	self[ #self + 1 ] = Texture;
	return Texture;
end
do
	local ApplyTransform;
	do
		local Det, AF, BF, CD, CE;
		local ULx, ULy, LLx, LLy, URx, URy, LRx, LRy;
		--- Applies an affine transformation to Texture.
		-- @param Texture  Texture to set TexCoords for.
		-- @param ...  First 6 elements of transformation matrix.
		function ApplyTransform( Texture, A, B, C, D, E, F )
			Det = A * E - B * D;
			if ( Det == 0 ) then
				return Texture:Hide(); -- Scaled infinitely small
			end
			AF, BF, CD, CE = A * F, B * F, C * D, C * E;

			ULx, ULy = ( BF - CE ) / Det, ( CD - AF ) / Det;
			LLx, LLy = ( BF - CE - B ) / Det, ( CD - AF + A ) / Det;
			URx, URy = ( BF - CE + E ) / Det, ( CD - AF - D ) / Det;
			LRx, LRy = ( BF - CE + E - B ) / Det, ( CD - AF - D + A ) / Det;

			-- Bounds to prevent "TexCoord out of range" errors
			if ( ULx < -1e4 ) then ULx = -1e4; elseif ( ULx > 1e4 ) then ULx = 1e4; end
			if ( ULy < -1e4 ) then ULy = -1e4; elseif ( ULy > 1e4 ) then ULy = 1e4; end
			if ( LLx < -1e4 ) then LLx = -1e4; elseif ( LLx > 1e4 ) then LLx = 1e4; end
			if ( LLy < -1e4 ) then LLy = -1e4; elseif ( LLy > 1e4 ) then LLy = 1e4; end
			if ( URx < -1e4 ) then URx = -1e4; elseif ( URx > 1e4 ) then URx = 1e4; end
			if ( URy < -1e4 ) then URy = -1e4; elseif ( URy > 1e4 ) then URy = 1e4; end
			if ( LRx < -1e4 ) then LRx = -1e4; elseif ( LRx > 1e4 ) then LRx = 1e4; end
			if ( LRy < -1e4 ) then LRy = -1e4; elseif ( LRy > 1e4 ) then LRy = 1e4; end

			return Texture:SetTexCoord( ULx, ULy, LLx, LLy, URx, URy, LRx, LRy );
		end
	end
	local min, max = min, max;
	local MinX, MinY, WindowX, WindowY;
	local ABx, ABy, BCx, BCy;
	local ScaleX, ScaleY, ShearFactor, Sin, Cos;
	local Parent, Width, Height;
	local SinScaleX, SinScaleY, CosScaleX, CosScaleY;
	local BorderScale, BorderOffset = 512 / 510, -1 / 512; -- Removes one-pixel transparent border
	local TrianglePath = [[Interface\AddOns\]]..ADDON_NAME..[[\Skin\Triangle]]
	--- Draw a triangle texture with vertices at relative coords.  (0,0) is top-left, (1,1) is bottom-right.
	function private:TextureAdd ( Layer, R, G, B, Alpha, Ax, Ay, Bx, By, Cx, Cy )
		ABx, ABy, BCx, BCy = Ax - Bx, Ay - By, Bx - Cx, By - Cy;
		ScaleX = ( BCx * BCx + BCy * BCy ) ^ 0.5;
		if ( ScaleX == 0 ) then
			return; -- Points B and C are the same
		end
		ScaleY = ( ABx * BCy - ABy * BCx ) / ScaleX;
		if ( ScaleY == 0 ) then
			return; -- Points are co-linear
		end
		ShearFactor = -( ABx * BCx + ABy * BCy ) / ( ScaleX * ScaleX );
		Sin, Cos = BCy / ScaleX, -BCx / ScaleX;

		-- Note: The texture region is made as small as possible to improve framerates.
		MinX, MinY = min( Ax, Bx, Cx ), min( Ay, By, Cy );
		WindowX, WindowY = max( Ax, Bx, Cx ) - MinX, max( Ay, By, Cy ) - MinY;


		-- Get a texture
		local Texture = private.TextureCreate( self, Layer, R, G, B, Alpha );
		Texture:SetTexture( TrianglePath );
		Width, Height = self:GetSize();
		Texture:SetPoint( "TOPLEFT", MinX * Width, -MinY * Height );
		Texture:SetSize( WindowX * Width, WindowY * Height );


		--[[ Transform parallelogram so its corners lie on the tri's points:
		local Matrix = Identity;
		-- Remove transparent border
		Matrix = Matrix:Scale( BorderScale, BorderScale );
		Matrix = Matrix:Translate( BorderOffset, BorderOffset );

		Matrix = Matrix:Shear( ShearFactor ); -- Shear the image so its bottom left corner aligns with point A
		Matrix = Matrix:Scale( ScaleX, ScaleY ); -- Scale X by the length of line BC, and Y by the length of the perpendicular line from BC to point A
		Matrix = Matrix:Rotate( Sin, Cos ); -- Align the top of the triangle with line BC.

		Matrix = Matrix:Translate( Bx - MinX, By - MinY ); -- Move origin to overlap point B
		Matrix = Matrix:Scale( 1 / WindowX, 1 / WindowY ); -- Adjust for change in texture size

		ApplyTransform( unpack( Matrix, 1, 6 ) );
		]]

		-- Common operations
		WindowX, WindowY = BorderScale / WindowX, BorderScale / WindowY;
		SinScaleX, SinScaleY = -Sin * ScaleX, Sin * ScaleY;
		CosScaleX, CosScaleY =  Cos * ScaleX, Cos * ScaleY;

		return ApplyTransform( Texture,
			WindowX * CosScaleX,
			WindowX * ( SinScaleY + CosScaleX * ShearFactor ),
			WindowX * ( ( SinScaleY + CosScaleX * ( 1 + ShearFactor ) ) * BorderOffset + Bx - MinX ) / BorderScale,
			WindowY * SinScaleX,
			WindowY * ( CosScaleY + SinScaleX * ShearFactor ),
			WindowY * ( ( CosScaleY + SinScaleX * ( 1 + ShearFactor ) ) * BorderOffset + By - MinY ) / BorderScale );
	end
end
--- Recycles all textures added to a frame using TextureCreate.
function private:TextureRemoveAll ()
	for Index = #self, 1, -1 do
		local Texture = self[ Index ];
		self[ Index ] = nil;
		Texture:Hide();
		Texture:SetParent( TexturesUnused );
		TexturesUnused[ #TexturesUnused + 1 ] = Texture;
	end
end


do
	local DWORD_LENGTH = 4;
	--- @return DWORD read from binary Data at Offset.
	local function ReadDWord ( Data, Offset )
		local B1, B2, B3, B4 = Data:byte( Offset + 1, Offset + DWORD_LENGTH );
		return B1 * 2 ^ 24 + B2 * 2 ^ 16 + B3 * 2 ^ 8 + B4; -- Big-endian
	end
	--- @return Offset of points, lines, and triangles within PathData.
	function private.GetPathPrimitiveOffsets ( PathData )
		return ReadDWord( PathData, 0 ) + 1, -- Points
			ReadDWord( PathData, DWORD_LENGTH ) + 1, -- Lines
			ReadDWord( PathData, DWORD_LENGTH * 2 ) + 1; -- Triangles
	end
end
do
	local COORD_MAX = 2 ^ 16 - 1;
	local BYTES_PER_TRIANGLE = 2 * 2 * 3;
	local Ax1, Ax2, Ay1, Ay2, Bx1, Bx2, By1, By2, Cx1, Cx2, Cy1, Cy2;
	--- Draws the given NPC's path onto a frame.
	-- @param PathData  Binary path data string.
	--[[
	function private:DrawPath ( PathData, Layer, R, G, B )
		local PointsOffset, LinesOffset, TrianglesOffset = private.GetPathPrimitiveOffsets( PathData );
		for Index = TrianglesOffset, #PathData, BYTES_PER_TRIANGLE do
			Ax1, Ax2, Ay1, Ay2, Bx1, Bx2, By1, By2, Cx1, Cx2, Cy1, Cy2 = PathData:byte( Index, Index + BYTES_PER_TRIANGLE - 1 );
			private.TextureAdd( self, Layer, R, G, B, 1,
				( Ax1 * 256 + Ax2 ) / COORD_MAX, 1 - ( Ay1 * 256 + Ay2 ) / COORD_MAX,
				( Bx1 * 256 + Bx2 ) / COORD_MAX, 1 - ( By1 * 256 + By2 ) / COORD_MAX,
				( Cx1 * 256 + Cx2 ) / COORD_MAX, 1 - ( Cy1 * 256 + Cy2 ) / COORD_MAX );
		end
	end
	]]--
	function private:InitializePath()
		return "\0\0\0\12\0\0\0\12\0\0\0\12"
	end

	-- Coordinates are expected to be in relative format (0,0) is upper left, (1,1) is lower right
	function private.AddTriangleToPath(PathData, Ax, Ay, Bx, By, Cx, Cy )
		local binTriangle = ""

--		DebugPrint("Cy is " .. Cy )

--		DebugPrint("Ax is " .. Ax .. " Ay is " .. Ay )

		Ax = COORD_MAX * Ax
		Ay = COORD_MAX * (1-Ay)
		Bx = COORD_MAX * Bx
		By = COORD_MAX * (1-By)
		Cx = COORD_MAX * Cx
		Cy = COORD_MAX * (1-Cy)


		local v1x1, v1x2 = floor(Ax / 256), floor(Ax % 256)
		local v1y1, v1y2 = floor(Ay / 256), floor(Ay % 256)

--		DebugPrint("Ax is " .. Ax .. " Ay is " .. Ay )
--		DebugPrint("v1x1 is " .. v1x1 .. " v1x2 is " .. v1x2 )
--		DebugPrint("v1y1 is " .. v1y1 .. " v1y2 is " .. v1y2 )

		binTriangle = binTriangle .. string.char( floor(Ax / 256) )
		binTriangle = binTriangle .. string.char( floor(Ax % 256) )
		binTriangle = binTriangle .. string.char( floor(Ay / 256) )
		binTriangle = binTriangle .. string.char( floor(Ay % 256) )


		binTriangle = binTriangle .. string.char( floor(Bx / 256) )
		binTriangle = binTriangle .. string.char( floor(Bx % 256) )
		binTriangle = binTriangle .. string.char( floor(By / 256) )
		binTriangle = binTriangle .. string.char( floor(By % 256) )


		binTriangle = binTriangle .. string.char( floor(Cx / 256) )
		binTriangle = binTriangle .. string.char( floor(Cx % 256) )
		binTriangle = binTriangle .. string.char( floor(Cy / 256) )
		binTriangle = binTriangle .. string.char( floor(Cy % 256) )

		return PathData .. binTriangle
	end

	function private.dumpTestPath( PathData )
		local PointsOffset, LinesOffset, TrianglesOffset = private.GetPathPrimitiveOffsets( PathData );

--		DebugPrint( "PointsOffset=" .. PointsOffset .. ", LinesOffset=" .. LinesOffset .. ", TrianglesOffset=" .. TrianglesOffset )

		for Index = TrianglesOffset, #PathData, BYTES_PER_TRIANGLE do
			Ax1, Ax2, Ay1, Ay2, Bx1, Bx2, By1, By2, Cx1, Cx2, Cy1, Cy2 = PathData:byte( Index, Index + BYTES_PER_TRIANGLE - 1 );

			local Ax, Ay = ( Ax1 * 256 + Ax2 ) / COORD_MAX, 1 - ( Ay1 * 256 + Ay2 ) / COORD_MAX
			local Bx, By = ( Bx1 * 256 + Bx2 ) / COORD_MAX, 1 - ( By1 * 256 + By2 ) / COORD_MAX
			local Cx, Cy = ( Cx1 * 256 + Cx2 ) / COORD_MAX, 1 - ( Cy1 * 256 + Cy2 ) / COORD_MAX


--			DebugPrint( "Triangle #" .. Index .. "A=(" .. Ax .. "," .. Ay ..") B=(" .. Bx .. "," .. By ..") C=(" .. Cx .. "," .. Cy ..")" )
		end
	end

	local math_sin, math_cos = math.sin, math.cos

	-- Max Distance, Min Distance, Angle left, Angle Right, facing )
	function private.CreatePath( maxDist, minDist, angleLeft, angleRight, facing )
		local pd



		local MapSzX, MapSzY = private.GetMapSize( private.GetMapID( GetRealZoneText() ) )

--		DebugPrint("MapSzX is " .. MapSzX .. " MapSzY is " .. MapSzY )

		if MapSzX and MapSzY then
			local Ax,Ay,Bx,By,Cx,Cy

			pd = private:InitializePath()

			--[[
			_________________
			\      /|\      /
			 \ 1  / | \  4 /
			  \  / 2|3 \  /
			   \/___|___\/
			]]

--			DebugPrint("player_x is " .. player_x .. " player_y is " .. player_y  )
--			DebugPrint("angleLeft is " .. angleLeft .. " angleRight is " .. angleRight .. " facing is " .. facing )
--			DebugPrint("maxDist is " .. maxDist .. " minDist is " .. minDist )

--			DebugPrint("cos(facing+angleLeft) is " .. math_cos(facing+angleLeft) )

			-- Triangle 1
			Ax = player_x - (minDist*math_sin(facing+angleLeft)) / MapSzX
			Ay = player_y - (minDist*math_cos(facing+angleLeft)) / MapSzY

			if ( Ax < 0 ) then Ax = 0 end
			if ( Ay < 0 ) then Ay = 0 end
			if ( Ax > 1 ) then Ax = 1 end
			if ( Ay > 1 ) then Ay = 1 end

			Bx = player_x - (maxDist*math_sin(facing)) / MapSzX
			By = player_y - (maxDist*math_cos(facing)) / MapSzY

			if ( Bx < 0 ) then Bx = 0 end
			if ( By < 0 ) then By = 0 end
			if ( Bx > 1 ) then Bx = 1 end
			if ( By > 1 ) then By = 1 end

			Cx = player_x - (maxDist*math_sin(facing+angleLeft)) / MapSzX
			Cy = player_y - (maxDist*math_cos(facing+angleLeft)) / MapSzY

			if ( Cx < 0 ) then Cx = 0 end
			if ( Cy < 0 ) then Cy = 0 end
			if ( Cx > 1 ) then Cx = 1 end
			if ( Cy > 1 ) then Cy = 1 end

			pd = private.AddTriangleToPath(pd, Ax, Ay, Bx, By, Cx, Cy )

			-- Triangle 2 shares vertices A and B with T1
			Cx = player_x - (minDist*math_sin(facing)) / MapSzX
			Cy = player_y - (minDist*math_cos(facing)) / MapSzY

			if ( Cx < 0 ) then Cx = 0 end
			if ( Cy < 0 ) then Cy = 0 end
			if ( Cx > 1 ) then Cx = 1 end
			if ( Cy > 1 ) then Cy = 1 end

			pd = private.AddTriangleToPath(pd, Ax, Ay, Bx, By, Cx, Cy )

			-- Triangle 3 shares vertices B and C with T2
			Ax = player_x - (minDist*math_sin(facing+angleRight-twoPi)) / MapSzX
			Ay = player_y - (minDist*math_cos(facing+angleRight-twoPi)) / MapSzY

			if ( Ax < 0 ) then Ax = 0 end
			if ( Ay < 0 ) then Ay = 0 end
			if ( Ax > 1 ) then Ax = 1 end
			if ( Ay > 1 ) then Ay = 1 end

			pd = private.AddTriangleToPath(pd, Ax, Ay, Bx, By, Cx, Cy )

			-- Finally, Triangle 4 shares vertices A and B with T3
			Cx = player_x - (maxDist*math_sin(facing+angleRight-twoPi)) / MapSzX
			Cy = player_y - (maxDist*math_cos(facing+angleRight-twoPi)) / MapSzY

			if ( Cx < 0 ) then Cx = 0 end
			if ( Cy < 0 ) then Cy = 0 end
			if ( Cx > 1 ) then Cx = 1 end
			if ( Cy > 1 ) then Cy = 1 end

			pd = private.AddTriangleToPath(pd, Ax, Ay, Bx, By, Cx, Cy )
--		else
--			DebugPrint( "Map Size was invalid." )
		end

		--private.dumpTestPath( pd )

		return pd



	end


end

--- Passes info for all enabled NPCs in a zone to a callback function.
-- @param Callback  Function( self, PathData, R, G, B )
function private:ApplyZone ( Map, Callback )

	local indeces = {
		[SURVEY_NONE] = 1,
		[SURVEY_GREEN] = 1,
		[SURVEY_YELLOW] = 1,
		[SURVEY_RED] = 1,
		}


	for _, surveyData in pairs(private.survey_info_this_dig) do

		-- table.insert( private.survey_info_this_dig, {x=player_x, y=player_y, map=player_map, level=player_level, color=SURVEY_RED, mindist=SURVEY_YARDS[SURVEY_YELLOW], maxdist=SURVEY_YARDS[SURVEY_RED], direction=facing, path=pd } )

		local colorTable = OVERLAY_SURVEY_COLORS[surveyData.color]

		if ( colorTable ) then
			local MapData = surveyData.path
			if ( MapData ) then

				local red,grn,blu = unpack(colorTable[indeces[surveyData.color]])

				-- DebugPrint("r=" .. red .. " g=" .. grn .. " b=" ..blu )

				Callback( self, MapData, unpack(colorTable[indeces[surveyData.color]]) );

				if ( indeces[surveyData.color] < #OVERLAY_SURVEY_COLORS[surveyData.color] ) then
					indeces[surveyData.color] = indeces[surveyData.color] + 1
				end

			end
		end
	end
end


----------------------------------------
----------------------------------------
fetchArchySurveyPOI = function()

	local kids = { Minimap:GetChildren() }
	local archyKids = {}

	for _, cFrame in pairs(kids) do
		local fName = cFrame:GetName()

		if fName and fName:match("ArchyMinimap_SurveyPOI%d+") then
			if cFrame.type == "survey" then
				table.insert(archyKids,cFrame)
			end
		end

	end

	return archyKids

end




----------------------------------------
----------------------------------------
local function enableAllArchySurveyPOI()

	if private.haveHiddenArchyIndicators then
		local kids = fetchArchySurveyPOI()

		for _, cFrame in pairs(kids) do
			-- If archy got called before us, then it will have removed the siteid field from the POI
			-- in the case that the active dig site we were in is no longer valid (we just surveyed the last one in this site)
			if cFrame.siteId then
				cFrame.icon:Show()
				cFrame:Show()
			end
		end

		private.haveHiddenArchyIndicators=nil

	end
end



----------------------------------------
-- minDist  POI closer to player than minDist will be hidden
-- maxDist  POI farther from player than maxDist will be hidden
----------------------------------------
local function hideArchySurveyPOI(minDist, maxDist, minAngle, maxAngle)

	local kids = fetchArchySurveyPOI()

	local playerFacing = _G.GetPlayerFacing()
	local rad_360 = math.rad(360)  -- use this to normalize resulting angle

	for _, cFrame in pairs(kids) do
		-- This gives the angle, starting from North, moving CClkwise from the player point (extended to North) to the given icon
		local angle = Astrolabe:GetDirectionToIcon( cFrame )
		if angle then
			-- To adjust for angle RELATIVE TO PLAYER, we need to SUBTRACT the player facing
			angle = angle - playerFacing
			if angle < 0 then
				angle = angle + rad_360
			end

			local showIt = false

			if angle <= minAngle or angle >= maxAngle then
				-- ok, it's in the right set of angles, now look for distance

				local dist = Astrolabe:GetDistanceToIcon( cFrame )

				if dist >= minDist and dist <= maxDist then
					showIt = true
				end


			end

			-- Do the arrow texture as well?
			if showIt then
				cFrame.icon:Show()
				cFrame:Show()
			else
				cFrame.icon:Hide()
				cFrame:Hide()

				private.haveHiddenArchyIndicators=true
			end

		end
	end


end


----------------------------------------
----------------------------------------
local function fetchArchyDigSiteFrame()

	local kids = { _G.UIParent:GetChildren() }
	local retval

	-- DebugPrint( "Looking for ArchyDigSiteFrame.  Found " .. #kids .. " child frames" )

	for _, cFrame in pairs(kids) do
		local fName = cFrame:GetName()

		if fName and fName == "ArchyDigSiteFrame" then
			-- DebugPrint( "Found ArchyDigSiteFrame" )
			retval = cFrame
			break
		end
	end

	return retval

end





----------------------------------------
----------------------------------------
local function userIsOnArchyNode()

-- Figure out siteID
-- Figure out map, floor, x and y
	local exists = false

	if not private.canDisplay or not private.archyFound then
		return
	end

	-- If there isn't a nearest indexer, then, well..
	if not nearestDigsiteIndexer then
--		DebugPrint("nearestDigsiteIndexer is nil")
		return exists
	end


	local map, level, x, y = Astrolabe:GetCurrentPlayerPosition()

	local siteId = nearestDigsiteIndexer.id

	if Archy.db.global.surveyNodes[siteId] then
		for _, node in pairs(Archy.db.global.surveyNodes[siteId]) do
			local distance = Astrolabe:ComputeDistance(map, level, x, y, node.m, node.f, node.x, node.y)

			if not distance or _G.IsInInstance() then
				distance = 0
			end

			if distance <= 10 then
--				DebugPrint("found a match in nearestDigsiteIndexer!")
				exists = true
				break
			end
		end
	end

	-- If there isn't a next nearest digsite, then return whatever we found with the nearest digsite
	-- Likewise, if we found something in the nearest digsite, then there's no way we could find something in the next nearest
	if (not nextNearestDigsiteIndexer) or (exists) then
--		DebugPrint("nextNearestDigsiteIndexer is nil or we already found one")
		return exists
	end


	siteId = nextNearestDigsiteIndexer.id

	if not Archy.db.global.surveyNodes[siteId] then
		--DebugPrint( "Couldn't find an Archy digsite for siteId " .. siteId )
		return exists
	end

	for _, node in pairs(Archy.db.global.surveyNodes[siteId]) do
		local distance = Astrolabe:ComputeDistance(map, level, x, y, node.m, node.f, node.x, node.y)

		if not distance or _G.IsInInstance() then
			distance = 0
		end

		if distance <= 10 then
--			DebugPrint("found a match in nextNearestDigsiteIndexer!")
			exists = true
			break
		end
	end

	return exists

end


----------------------------------------
-- Compare where I am now in relation to the last survey.  If it's close enough (say 4 yards), then if the user pressed a distance
-- button, we're close enough to remember that button press.
----------------------------------------
function closeToLastSurvey()

	local p_map, p_level, p_x, p_y = Astrolabe:GetCurrentPlayerPosition()

	local distance = Astrolabe:ComputeDistance(p_map, p_level, p_x, p_y, private.survey_map, private.survey_level, private.survey_x, private.survey_y)

	if not distance then
		distance=1000
	end

	if distance < 5 then
		return true
	end

	return false

end

----------------------------------------
----------------------------------------
function buttonRed_OnLClick()

	DebugPrint( "buttonRed_OnLClick" )

	if closeToLastSurvey() and private.enableSurveyRecall then
		local facing =_G.GetPlayerFacing()

		-- Max Distance, Min Distance, Angle left, Angle Right, facing )
		local pd = private.CreatePath( SURVEY_YARDS[SURVEY_RED], SURVEY_YARDS[SURVEY_YELLOW], SURVEY_ANGLES[SURVEY_RED].left, SURVEY_ANGLES[SURVEY_RED].right, facing )

		hideArchySurveyPOI( SURVEY_YARDS[SURVEY_YELLOW], SURVEY_YARDS[SURVEY_RED], SURVEY_ANGLES[SURVEY_RED].left, SURVEY_ANGLES[SURVEY_RED].right )

		table.insert( private.survey_info_this_dig, {x=player_x, y=player_y, map=player_map, level=player_level, color=SURVEY_RED, mindist=SURVEY_YARDS[SURVEY_YELLOW], maxdist=SURVEY_YARDS[SURVEY_RED], direction=facing, path=pd } )

		private.Modules.UpdateMap(nil)
	end

end


----------------------------------------
----------------------------------------
function buttonYellow_OnLClick()

	DebugPrint( "buttonYellow_OnLClick" )

	if closeToLastSurvey() and private.enableSurveyRecall then

		local facing =_G.GetPlayerFacing()

		-- Min Distance, Max Distance, Angle left, Angle Right, facing )
		local pd = private.CreatePath( SURVEY_YARDS[SURVEY_YELLOW], SURVEY_YARDS[SURVEY_GREEN], SURVEY_ANGLES[SURVEY_YELLOW].left, SURVEY_ANGLES[SURVEY_YELLOW].right, facing )

		hideArchySurveyPOI( SURVEY_YARDS[SURVEY_GREEN], SURVEY_YARDS[SURVEY_YELLOW], SURVEY_ANGLES[SURVEY_YELLOW].left, SURVEY_ANGLES[SURVEY_YELLOW].right )

		table.insert( private.survey_info_this_dig, {x=player_x, y=player_y, map=player_map, level=player_level, color=SURVEY_YELLOW, mindist=SURVEY_YARDS[SURVEY_GREEN], maxdist=SURVEY_YARDS[SURVEY_YELLOW], direction=facing, path=pd } )

		private.Modules.UpdateMap(nil)
	end

end

----------------------------------------
----------------------------------------
function buttonGreen_OnLClick()

	DebugPrint( "buttonGreen_OnLClick" )

	if closeToLastSurvey() and private.enableSurveyRecall then

		local facing =_G.GetPlayerFacing()

		-- Min Distance, Max Distance, Angle left, Angle Right, facing )
		local pd = private.CreatePath( SURVEY_YARDS[SURVEY_GREEN], MIN_DRY_SURVEY_DISTANCE, SURVEY_ANGLES[SURVEY_GREEN].left, SURVEY_ANGLES[SURVEY_GREEN].right, facing )

		hideArchySurveyPOI( MIN_DRY_SURVEY_DISTANCE, SURVEY_YARDS[SURVEY_GREEN], SURVEY_ANGLES[SURVEY_GREEN].left, SURVEY_ANGLES[SURVEY_GREEN].right )

		table.insert( private.survey_info_this_dig, {x=player_x, y=player_y, map=player_map, level=player_level, color=SURVEY_GREEN, mindist=MIN_DRY_SURVEY_DISTANCE, maxdist=SURVEY_YARDS[SURVEY_GREEN], direction=facing, path=pd } )

		private.Modules.UpdateMap(nil)
	end

end

----------------------------------------
----------------------------------------
function buttonReset_OnLClick()

	if private.db.general.useArchy then
		enableAllArchySurveyPOI()
	end

	private.survey_info_this_dig = {}

	private.Modules.UpdateMap(nil)

	private.enableSurveyRecall=nil
end


----------------------------------------
----------------------------------------
function Jughead:SaveFramePosition(frame)

	local bPoint, bRelativeTo, bRelativePoint, bXofs, bYofs = frame:GetPoint()
	local width, height
	local anchor, relativePoint
	local xOfs, yOfs

	if frame and frame == myButtonFrame and not _G.InCombatLockdown() then

--		DebugPrint( "In SaveFramePosition.  I found that my new anchor point is " .. bPoint .." and my relativeto anchor point is " .. bRelativePoint );

		-- First, translate back to my desired anchor point
		anchor = private.db.frame.anchor

		-- WoW chose my desired anchor point.  Woot!
		if anchor == bPoint then
			relativePoint = bRelativePoint
			xOfs = bXofs
			yOfs = bYofs
		else
			-- Translate my new anchor point back to TOPLEFT
			width = frame:GetWidth()
			height = frame:GetHeight()

			if bPoint == "TOP" then
				bXofs = bXofs - (width / 2)
			elseif bPoint == "LEFT" then
				bYofs = bYofs + (height / 2)
			elseif bPoint == "BOTTOMLEFT" then
				bYofs = bYofs + height
			elseif bPoint == "TOPRIGHT" then
				bXofs = bXofs - width
			elseif bPoint == "RIGHT" then
				bYofs = bYofs + (height / 2)
				bXofs = bXofs - width
			elseif bPoint == "BOTTOM" then
				bYofs = bYofs + height
				bXofs = bXofs - (width / 2)
			elseif bPoint == "BOTTOMRIGHT" then
				bYofs = bYofs + height
				bXofs = bXofs - width
			elseif bPoint == "CENTER" then
				bYofs = bYofs + (height / 2)
				bXofs = bXofs - (width / 2)
			end

			relativePoint = bRelativePoint
			xOfs = bXofs
			yOfs = bYofs
		end

		private.db.frame.relativeTo = bRelativePoint;
		private.db.frame.positionX = xOfs;
		private.db.frame.positionY = yOfs;


		frame:SetUserPlaced(false)
		frame:ClearAllPoints()
		frame:SetPoint("TOPLEFT", _G.UIParent, private.db.frame.relativeTo, private.db.frame.positionX, private.db.frame.positionY)
	end
end


local DIG_LOCATION_TEXTURE = 177
----------------------------------------
-- Called by UpdatePlayerPosition whenever continent_digsites[current_continent] is empty or nil
-- This happens on initial load and when we receive the ARTIFACT_DIG_SITE_UPDATED event or when the player changes continents
-- UpdatePlayerPosition only calls this when the world map frame is closed
----------------------------------------
local function GetContinentSites(continent_id)

	--DebugPrint( "In GetContinentSites" )

	local new_sites = {}
	local showDig = _G.GetCVarBool("digSites")
	if not showDig then
		_G.SetCVar("digSites", "1")
		_G.WorldMapArchaeologyDigSites:Show()
		--_G.WorldMapShowDigSites:SetChecked(1)
		_G.RefreshWorldMap()
		showDig = "0"
	end

	if not continent_id then
		--DebugPrint( "continent_id was nil" )
		return new_sites
	end

	_G.SetMapZoom(continent_id)

	for index = 1, _G.GetNumMapLandmarks() do
		local name, description, texture_index, activeDigsiteX, activeDigsiteY = _G.GetMapLandmarkInfo(index)

		if texture_index == DIG_LOCATION_TEXTURE then
			local staticSiteData = private.dig_sites[name]
			continentMapID = Astrolabe:GetMapID(continent_id, 0)

			if staticSiteData then
				siteMapID = staticSiteData.map

				local worldX, worldY = Astrolabe:TranslateWorldMapPosition(continentMapID, 0, activeDigsiteX, activeDigsiteY, siteMapID, 0)

				local digsite = {
					continent = continentMapID,
					map = siteMapID,
					level = 0,
					x = worldX,
					y = worldY,
					name = name,
					id = staticSiteData.blob_id,  -- This is the indexer into the digsites location database
					distance = 999999
				}
				table.insert(new_sites, digsite)
			end
		end
	end
	if showDig == "0" then -- restore initial setting
		_G.SetCVar("digSites", showDig)
		_G.WorldMapArchaeologyDigSites:Hide()
		--_G.WorldMapShowDigSites:SetChecked(nil)
		_G.RefreshWorldMap()
	end
	return new_sites
end


----------------------------------------
----------------------------------------
local function SortSitesByDistance(a, b)
	return (a.distance or 0) < (b.distance or 0)
end


----------------------------------------
-- Called by UpdatePlayerPosition to find the nearest two digsites on the current continent
----------------------------------------
function setNearestDigsites()

	--DebugPrint( "In setNearestDigsites" )

	if not continent_digsites[current_continent] or (#continent_digsites[current_continent] == 0) then
		nearestDigsiteIndexer = nil
		nextNearestDigsiteIndexer = nil

		return
	end

	for index = 1, SITES_PER_CONTINENT do
		local site = continent_digsites[current_continent][index]

		site.distance = Astrolabe:ComputeDistance(player_map, player_level, player_x, player_y, site.map, site.level, site.x, site.y)
	end


	-- Sort sites
	local sites = continent_digsites[current_continent]
	table.sort(sites, SortSitesByDistance)

	local nindex = nil
	nindex, nearestDigsiteIndexer = next(sites,nil)
	_, nextNearestDigsiteIndexer = next(sites,nindex)
end



----------------------------------------
----------------------------------------
local function SetDistanceIndicatorColor(color)

	if private.frameCreated then
--		myButtonFrame.distind:Hide()

		if color == SURVEY_NONE or #private.survey_info_this_dig == 0 then
			myButtonFrame.distind:Hide()

			if timer_handle then
				Jughead:CancelTimer(timer_handle, true)

				timer_handle = Jughead:ScheduleRepeatingTimer("UpdatePlayerPosition", 2)
			end
		else
			myButtonFrame.distind.texture:SetTexCoord(unpack(DISTANCE_COLOR_TEXCOORDS[color]))
			-- myButtonFrame.distind:SetAlpha(1)
			myButtonFrame.distind:Show()

			-- Update more often since we have to update the distance indicator
			if timer_handle then
				Jughead:CancelTimer(timer_handle, true)

				timer_handle = Jughead:ScheduleRepeatingTimer("UpdatePlayerPosition", .5)
			end
		end
	end
end

----------------------------------------
----------------------------------------
local function UpdateDistanceIndicator()
	if _G.IsInInstance() then
		return
	end

	local finalColor=SURVEY_NONE

	for _, digloc in pairs(private.survey_info_this_dig) do
		local distance = Astrolabe:ComputeDistance(player_map, player_level, player_x, player_y, digloc.map, digloc.level, digloc.x, digloc.y)

		if not distance then
			distance = 0
		end

		if digloc.color == SURVEY_GREEN then
			if distance >= MIN_DRY_SURVEY_DISTANCE and distance <= SURVEY_YARDS[SURVEY_GREEN] then
				if finalColor < SURVEY_GREEN then
					finalColor = SURVEY_GREEN
				end
			elseif distance <= SURVEY_YARDS[SURVEY_GREEN]+5 then
				if finalColor < SURVEY_YELLOW then
					finalColor = SURVEY_YELLOW
				end
			else
				finalColor = SURVEY_RED
			end
		elseif digloc.color == SURVEY_YELLOW then
			if distance >= SURVEY_YARDS[SURVEY_GREEN] and distance <= SURVEY_YARDS[SURVEY_YELLOW] then
				if finalColor < SURVEY_GREEN then
					finalColor = SURVEY_GREEN
				end
			elseif distance >= SURVEY_YARDS[SURVEY_YELLOW] and distance <= SURVEY_YARDS[SURVEY_YELLOW]+10 then
				if finalColor < SURVEY_YELLOW then
					-- just beyond the yellow zone?
					finalColor = SURVEY_YELLOW
				end
			elseif distance >= SURVEY_YARDS[SURVEY_YELLOW]-5 and distance <= SURVEY_YARDS[SURVEY_YELLOW] then
				if finalColor < SURVEY_YELLOW then
					-- just inside the green zone?
					finalColor = SURVEY_YELLOW
				end
			else
				finalColor = SURVEY_RED
			end
		elseif digloc.color == SURVEY_RED then
			if distance >= SURVEY_YARDS[SURVEY_YELLOW] and distance <= SURVEY_YARDS[SURVEY_RED] then
				if finalColor < SURVEY_GREEN then
					finalColor = SURVEY_GREEN
				end
			elseif distance >= SURVEY_YARDS[SURVEY_RED]-20 then
				if finalColor < SURVEY_YELLOW then
					-- just inside the yellow zone?
					finalColor = SURVEY_YELLOW
				end
			else
				finalColor = SURVEY_RED
			end
		end

	end


	SetDistanceIndicatorColor(finalColor)

end


----------------------------------------
--[[ Survey Functions ]] --
-- Copied from Archy to interface with the Archy survey database
----------------------------------------
local function AddSurveyNode(siteId, map, level, x, y)
	local newNode = {
		m = map,
		f = level,
		x = x,
		y = y
	}
	local exists = false

	if not Archy.db.global.surveyNodes then
		Archy.db.global.surveyNodes = {}
	end

	if not Archy.db.global.surveyNodes[siteId] then
		Archy.db.global.surveyNodes[siteId] = {}
	end

	for _, node in pairs(Archy.db.global.surveyNodes[siteId]) do
		local distance = Astrolabe:ComputeDistance(newNode.m, newNode.f, newNode.x, newNode.y, node.m, node.f, node.x, node.y)

		if not distance or _G.IsInInstance() then
			distance = 0
		end

		if distance <= 10 then
			exists = true
			break
		end
	end
	if not exists then
		table.insert(Archy.db.global.surveyNodes[siteId], newNode)
	end
end


-----------------------------------------------------------------------
--
-- Event handlers and timers
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Mouse events
-----------------------------------------------------------------------

----------------------------------------
----------------------------------------
function Jughead_ButtonRed_OnMouseDown(self, button)
--	if button == "LeftButton" then
		buttonRed_OnLClick()
--	elseif button == "RightButton" then
--		buttonRed_OnRClick()
--	end
end

----------------------------------------
----------------------------------------
function Jughead_ButtonYellow_OnMouseDown(self, button)
	buttonYellow_OnLClick()
end

----------------------------------------
----------------------------------------
function Jughead_ButtonGreen_OnMouseDown(self, button)
	buttonGreen_OnLClick()
end

----------------------------------------
----------------------------------------
function Jughead_ButtonReset_OnMouseDown(self, button)
	buttonReset_OnLClick()
end


-----------------------------------------------------------------------
-- Timers
-----------------------------------------------------------------------

----------------------------------------
-- Called by the timer we set up in PLAYER_ENTERING_WORLD event
----------------------------------------
function Jughead:UpdatePlayerPosition(force)

	if _G.IsInInstance() or _G.UnitIsGhost("player") then
		return
	end

	player_map, player_level, player_x, player_y = Astrolabe:GetCurrentPlayerPosition()
	UpdateDistanceIndicator()

	-- Please note that GetCurrentMapContinent shows the currently displayed continent on the world map.
	-- This is NOT NECESSARILY where the player is, if the player has the world map open.
	-- Therefore, if the world map is open, don't update current_continent...
	if WorldMapFrame:IsVisible() then
		return
	end

	if _G.GetCurrentMapAreaID() == -1 then
		return
	end

	current_continent = _G.GetCurrentMapContinent()


	-- These functions will change the position of the world map, so save where we are now
	local original_map_id = _G.GetCurrentMapAreaID()

	-- Update the list of active digsites on this continent, if necessary
	if not continent_digsites then
		continent_digsites = {}
	end

	if not continent_digsites[current_continent] or (#continent_digsites[current_continent] == 0) then
		continent_digsites[current_continent] = GetContinentSites(current_continent)

		--UpdateSite(current_continent)
	end
	_G.SetMapByID(original_map_id)

	if not player_map or not player_level or (player_x == 0 and player_y == 0) then
		return
	end

	setNearestDigsites()

end


-----------------------------------------------------------------------
-- Game events
-----------------------------------------------------------------------

----------------------------------------
-- When we enter the world, start keeping track of where the player is
----------------------------------------
function Jughead:PLAYER_ENTERING_WORLD()
	-- Two timers are needed here: If we force a call to UpdatePlayerPosition() too soon, the site distances will not update properly and the notifications may vanish just as the player is able to see them.
	if not timer_handle and private.canDisplay then
		self:ScheduleTimer(function()
			timer_handle = self:ScheduleRepeatingTimer("UpdatePlayerPosition", 2)
		end, 2)
	end

end


----------------------------------------
-- Entering combat
----------------------------------------
function Jughead:PLAYER_REGEN_DISABLED()
	private.in_combat = true
end

----------------------------------------
-- Leaving combat
----------------------------------------
function Jughead:PLAYER_REGEN_ENABLED()

	private.in_combat = nil

end



----------------------------------------
----------------------------------------
function Jughead:CURRENCY_DISPLAY_UPDATE()

	if not current_continent or _G.GetNumArchaeologyRaces() == 0 or not private.canDisplay then
		return
	end

	if not private.db.general.enabled then
		return
	end

	local goodDig = false
	for race_id = 1, _G.GetNumArchaeologyRaces() do
		local _, _, _, currencyAmount = _G.GetArchaeologyRaceInfo(race_id)
		if ( not race_data[race_id] ) then
			race_data[race_id] = {}
		end
		local diff = currencyAmount - (race_data[race_id].currency or 0)

		race_data[race_id].currency = currencyAmount

		if diff < 0 then
			-- we've "spent" fragments by solving an artifact.. nothing to do here, move along..
		elseif diff > 0 then
			-- we've gained fragments so we have successfully dug.  Turn the survey nodes back on

			goodDig = true

			private.survey_info_this_dig = {}

			private.Modules.UpdateMap(nil)
		end
	end

	-- If we found something, and Archy is enabled, then turn the archy pins back on
	if goodDig and private.archyFound and private.db.general.useArchy and not private.lastDigInSiteSurveyed then
		enableAllArchySurveyPOI()
	end

	private.lastDigInSiteSurveyed = false


	-- If we found something, and both GM2 and Archy are enabled, see if we can copy any nodes from the
	-- GM2 database to the Archy database.
	if goodDig and private.archyFound and private.db.general.useArchy and private.gm2Found and nearestDigsiteIndexer then

		local x, y = GetPlayerMapPosition("player")
		local zone = GetCurrentMapAreaID()
		local level = GetCurrentMapDungeonLevel()

		local numFound = 0
		for coord, nodeID in GatherMate:FindNearbyNode(zone, x, y, level, "Archaeology", 200, true) do
			numFound = numFound + 1

			-- Decode GM2 x,y,level
			local x2, y2, level2 = floor(coord/1000000)/10000, floor(coord % 1000000 / 100)/10000, coord % 100

--			DebugPrint( "(" .. x2 .. "," .. y2 .. "," .. level2 .. ")")

			AddSurveyNode(nearestDigsiteIndexer.id, nearestDigsiteIndexer.map, level2, x2, y2)

		end

--		DebugPrint( "Found " .. numFound .. " GM2 nodes nearby.")
	end

end


----------------------------------------
----------------------------------------
function Jughead:ARTIFACT_DIG_SITE_UPDATED()
	if not current_continent then
		return
	end

	-- This will cause UpdatePlayerPosition to requery sites and recompute distances
	continent_digsites[current_continent] = nil

end


----------------------------------------
----------------------------------------
function Jughead:ConfigUpdated()

	-- We weren't enabled at addon load, so we didn't create our frame, but now we're enabled
	if not private.frameCreated and private.db.general.enabled then
		createFrame(_G.UIParent)

		Jughead:RegisterChatCommand("jughead", "MySlashProcessorFunc")

		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
		self:RegisterEvent("ARTIFACT_DIG_SITE_UPDATED")
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")

	end

	if private.frameCreated then

		local enableStateChanged = false

		-- Did the user change enabled state on us?
		if private.canDisplay and not private.db.general.enabled then
			private.canDisplay = false
			enableStateChanged = true
		elseif not private.canDisplay and private.db.general.enabled then
			private.canDisplay = true
			enableStateChanged = true
		end


		myButtonFrame:SetScale(private.db.frame.scale)


		if private.db.general.enabled and private.canDisplay then
			if private.db.frame.visible then
				myButtonFrame:Show()
			else
				myButtonFrame:Hide()
			end
		else
			enableAllArchySurveyPOI() -- reenable all the POIs so that there won't be any strangeness when Archy refreshes the minimap in CURRENCY_DISPLAY_UPDATE

			private.enableSurveyRecall = nil

--			DebugPrint( "ConfigUpdated private.enableSurveyRecall unset" )

			private.survey_info_this_dig={}

			private.Modules.UpdateMap(nil)

			myButtonFrame:Hide()
		end
	end

end

----------------------------------------
----------------------------------------
function Jughead:OnProfileUpdate(event, database, ProfileKey)
	private.db = database and database.profile or self.db.profile

	if private.frameCreated then
		self:ConfigUpdated()
	end
end

SURVEY_SPELL_ID=80451
----------------------------------------
----------------------------------------
function Jughead:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell, rank, line_id, spell_id)
	if unit ~= "player" or spell_id ~= SURVEY_SPELL_ID then
--		DebugPrint( "Wasn't spell" )
		return
	end

	if not private.db.general.enabled then
--		DebugPrint( "Wasn't enabled" )
		return
	end

	if private.db.general.autoshow then
		Jughead:ShowFrame()
	end

	-- Remember where we dug, so that (if) the user presses a distance button, we can record it
	private.survey_map, private.survey_level, private.survey_x, private.survey_y = Astrolabe:GetCurrentPlayerPosition()
	private.enableSurveyRecall = true  -- we just dug, so if the user presses one of the distance buttons w/o moving too far, record it.

	--DebugPrint( "private.enableSurveyRecall set" )

	-- Look for the condition where the player just surveyed and found something.  At that point, the digsites have
	-- changed, and we want to avoid leaving artifacts on the minimap from the old digsite
	local lastNearestDigsiteIndexer = nearestDigsiteIndexer
	self.UpdatePlayerPosition()

	-- ok, digsites have changed..	That means that we just successfully dug the last site in this dig site.
	if lastNearestDigsiteIndexer ~= nearestDigsiteIndexer then
		private.lastDigInSiteSurveyed = true  -- leave a marker for CURRENCY_DISPLAY_UPDATE

		if private.db.general.useArchy then
			enableAllArchySurveyPOI() -- reenable all the POIs so that there won't be any strangeness when Archy refreshes the minimap in CURRENCY_DISPLAY_UPDATE
		end

		private.enableSurveyRecall = nil

		--DebugPrint( "UNIT_SPELLCAST_SUCCEEDED private.enableSurveyRecall unset" )

		private.survey_info_this_dig={}

		private.Modules.UpdateMap(nil)
	end


end

-----------------------------------------------------------------------
--
-- Addon events
--
-----------------------------------------------------------------------

----------------------------------------
----------------------------------------
function Jughead:OnInitialize()

	self.db = LibStub("AceDB-3.0"):New("JugheadDB", PROFILE_DEFAULTS, 'Default')
	self.db.RegisterCallback(self, "OnNewProfile", "OnProfileUpdate")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileUpdate")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileUpdate")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileUpdate")

	private.db = self.db.profile

end

local function frame_OnEnter(self)
	_G.GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT")
	_G.GameTooltip:SetText("Control-Left Click to toggle the frame lock\nAlt-Left Click to hide the frame", _G.NORMAL_FONT_COLOR[1], _G.NORMAL_FONT_COLOR[2], _G.NORMAL_FONT_COLOR[3], 1)
end

local function frame_OnLeave(self)
	_G.GameTooltip:Hide()
end


----------------------------------------
-- Called to create my frame.  If I'm attaching to
-- Archy, this is called in a timer to give Archy a chance
-- to get itself set up.
----------------------------------------
local function createFrame(parent)

	myButtonFrame = _G.CreateFrame("Frame", "JugheadDistanceAttachmentFrame", parent, "JugheadDistanceAttachment")
	private.frameCreated = true

	if private.db then
		myButtonFrame:ClearAllPoints()
		myButtonFrame:SetPoint("TOPLEFT", _G.UIParent, private.db.frame.relativeTo, private.db.frame.positionX, private.db.frame.positionY)
		myButtonFrame:SetScale(private.db.frame.scale)

		myButtonFrame:SetScript("OnEnter", frame_OnEnter)
		myButtonFrame:SetScript("OnLeave", frame_OnLeave)


	else
		myButtonFrame:SetScale(1.0)
	end

	myButtonFrame.distind:Hide()

	if private.db.frame.visible then
		myButtonFrame:Show()
	else
		myButtonFrame:Hide()
	end

end


----------------------------------------
-- Called when the addon is enabled
----------------------------------------
function Jughead:OnEnable()

	private.archyFound = false
	private.gm2Found = false
	private.canDisplay = false

	-- Belt and suspenders..
	if not self.db.profile then
		self.db.profile = PROFILE_DEFAULTS
	end
	if not self.db.profile.general then
		self.db.profile.general = PROFILE_DEFAULTS.general
	end


	-- Decide if I found my target addons, and if the user has asked me to interface with them
	if _G["Archy"] and self.db.profile.general.useArchy then
		private.archyFound = true
	end

	if GatherMate and self.db.profile.general.useGM2 then
		private.gm2Found = true
	end

	-- Keep the settings variable (db.general.enabled) separate from the "working" variable (canDisplay)
	-- so we can see changes during addon config.
	if  private.db.general.enabled then
		private.canDisplay = true
	end

	local ALS = LibStub("AceConfig-3.0")
	ALS:RegisterOptionsTable("Jughead", Jughead.options, "jh")
	private.Modules.RegisterOptions(ALS)
	--LibStub("AceConfig-3.0"):RegisterOptionsTable("JugheadMinimapMod", GetArtifactOptions)

	local ACD = LibStub("AceConfigDialog-3.0")
	private.optionsFrame = ACD:AddToBlizOptions("Jughead", "Jughead", nil)
	private.Modules.AddToBlizOptions(ACD,"Jughead")

	--LibStub("AceConfigDialog-3.0"):AddToBlizOptions("JugheadMinimapMod", "Minimap", "Jughead")


	-- If I found Archy, and the user wants me to use Archy, then hook myself up
	if private.db.general.enabled and private.archyFound then

		--DebugPrint( ADDON_NAME .. " " .. Jughead.version .. " " .. Archy.version )

		--local archyVers1, archyVers2, archyVers3 = strsplit(".", Archy.version)

		if Archy.db and Archy.db.global and Archy.db.global.surveyNodes then
			DebugPrint("I found the Archy survey database.")
		end

	end


	-- If I found GM2, and the user wants me to use it...
	if private.db.general.enabled and private.gm2Found then

		DebugPrint( "I found GM2." )

	end

	if private.db.general.enabled then

		createFrame(_G.UIParent)

		Jughead:RegisterChatCommand("jughead", "MySlashProcessorFunc")

		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
		self:RegisterEvent("ARTIFACT_DIG_SITE_UPDATED")
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")



	end
end

----------------------------------------
----------------------------------------
function Jughead:OnDisable()
    -- Called when the addon is disabled
    if private.canDisplay then
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("CURRENCY_DISPLAY_UPDATE")
		self:UnregisterEvent("ARTIFACT_DIG_SITE_UPDATED")
		self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")

    	if myButtonFrame then
    		myButtonFrame:Hide()
    	end
    end

    if timer_handle then
	    self:CancelTimer(timer_handle)
	end

end


----------------------------------------
----------------------------------------
dumpMinimapStats = function()


	local indeces = {
		[SURVEY_NONE] = 1,
		[SURVEY_GREEN] = 1,
		[SURVEY_YELLOW] = 1,
		[SURVEY_RED] = 1,
		}


	DebugPrint((private.UpdateCount or 0) .. " " .. (private.CancelCount or 0))

	DebugPrint((private.UpdateMap or -1) .. " " .. (private.CurrentMapID or -1))


	for _, surveyData in pairs(private.survey_info_this_dig) do

		local colorTable = OVERLAY_SURVEY_COLORS[surveyData.color]

		if ( colorTable ) then

			local MapData = surveyData.path
			if ( MapData ) then

				private.dumpTestPath(MapData)

				if ( indeces[surveyData.color] < #OVERLAY_SURVEY_COLORS[surveyData.color] ) then
					indeces[surveyData.color] = indeces[surveyData.color] + 1
				end

			end
		end
	end

end


----------------------------------------
----------------------------------------
function Jughead:HideFrame()
	if private.frameCreated then
		myButtonFrame:Hide()
		private.db.frame.visible = false
	end
end

----------------------------------------
----------------------------------------
function Jughead:ShowFrame()
	if private.frameCreated then
		myButtonFrame:Show()
		private.db.frame.visible = true
	end
end

----------------------------------------
-- Slash command handler
----------------------------------------
function Jughead:MySlashProcessorFunc(input)
  -- Process the slash command ('input' contains whatever follows the slash command)
    if private.canDisplay then
		local command = input:lower()

		if command == L["dump"]:lower() then
			dumpMinimapStats()

		elseif command == L["red"]:lower() then
			buttonRed_OnLClick()
		elseif command == L["yellow"]:lower() then
			buttonYellow_OnLClick()
		elseif command == L["green"]:lower() then
			buttonGreen_OnLClick()
		elseif command == L["SHOW"]:lower() then
			Jughead:ShowFrame()
		elseif command == L["HIDE"]:lower() then
			Jughead:HideFrame()
		elseif command == L["AUTOSHOW"]:lower() then
			Jughead:Print(L["Jughead will autoshow on dig."])
			private.db.general.autoshow = true
		elseif command == L["NOAUTOSHOW"]:lower() then
			Jughead:Print(L["Jughead will not autoshow on dig."])
			private.db.general.autoshow = false
		else
			Jughead:Print(L["Jughead commands:"])
			Jughead:Print(L["red"]:lower() .. ":" .. L["red_cmd_desc"])
			Jughead:Print(L["yellow"]:lower() .. ":" .. L["yellow_cmd_desc"])
			Jughead:Print(L["green"]:lower() .. ":" .. L["green_cmd_desc"])
			Jughead:Print(L["SHOW"]:lower() .. ":" .. L["show_cmd_desc"])
			Jughead:Print(L["HIDE"]:lower() .. ":" .. L["hide_cmd_desc"])
			Jughead:Print(L["AUTOSHOW"]:lower() .. ":" .. L["autoshow_cmd_desc"])
			Jughead:Print(L["NOAUTOSHOW"]:lower() .. ":" .. L["noautoshow_cmd_desc"])
		end

    end
end



--- Reloads enabled modules from saved settings.
function private.Synchronize ( Options )
	-- Load defaults if settings omitted
	if ( not Options ) then
		Options = private.OptionsDefault;
	end

	private.Modules.OnSynchronize( Options );
end

do
	local MapIDs = {}; -- [ LocalizedZoneName ] = MapID;
	local MapWidths, MapHeights = {}, {}; -- [MapID] = width,height

	function private.NewMap ( Name )
		if not MapIDs[ Name ] then

			local MapID = -1

			if ( Name ~= GetRealZoneText() )  then -- Asking for a map not the current one?
				DebugPrint("Not asking for current zone")
				return
			end

			-- Please note that GetCurrentMapContinent shows the currently displayed continent on the world map.
			-- This is NOT NECESSARILY where the player is, if the player has the world map open.
			-- Therefore, if the world map is open, don't update current_continent...
			if WorldMapFrame:IsVisible() then
				DebugPrint("World map is visible")
				return
			end

			local MapID = _G.GetCurrentMapAreaID()

			if MapID == -1 then
				--DebugPrint("MapID was -1")
				return
			end


			local _, X1, Y1, X2, Y2 = GetCurrentMapZone();

			local Width, Height = X1 - X2, Y1 - Y2;
			if ( Width == 0 or Height == 0 ) then
				--DebugPrint("Map width or height was 0")
				return
			end
			MapIDs[ Name ] = MapID

			MapWidths[ MapID ], MapHeights[ MapID ] = Width, Height;

		end
	end


	--- @return Map ID for localized zone name or nil if unknown.
	-- Note that only true continent sub-zones are supported.
	function private.GetMapID ( Name )
		--DebugPrint("In GetMapID")
		if not Name then
			--DebugPrint("GetMapID:Name was nil")
			return -1
		end

		if not MapIDs[ Name ] then
			-- Careful if in a dungeon..
			--DebugPrint("GetMapID:Calling NewMap.  Name was " .. Name)
			private.NewMap( Name )
			--DebugPrint("MapID[Name] is " .. MapIDs[ Name ])
		end
		return MapIDs[ Name ];
	end
	--- @return Width and height of Map in yards or nil if unavailable.
	function private.GetMapSize ( Map )
		local mapx, mapy = 0,0
		if not MapWidths[ Map ] then
			DebugPrint("Trying to create new map info in GetMapSize")

			local realZone = GetRealZoneText()

			if ( realZone ) then
				-- If we don't have the mapID of the current player zone, then we can't compare the requested Map with the current zone
				-- If we can't do that, then we can't tell if we can create the requested data
				-- If the user is asking for data on the current map, then we can do it
				if MapIDs[ realZone ] and MapIDs[ realZone ] == Map then
					private.NewMap( realZone )
				end
			end
		end

		if MapWidths[ Map ] then
			mapx, mapy = MapWidths[ Map ], MapHeights[ Map ]
		end

		return mapx, mapy
	end


	--- Loads defaults, validates settings, and begins listening for Overlay API messages.
	function Jughead:ADDON_LOADED ( Event, AddOn )
		if ( AddOn == ADDON_NAME ) then
			self[ Event ] = nil;
			self:UnregisterEvent( Event );

			local Options = JugheadDB.Options;
			JugheadDB.Options = private.Options;
			if ( Options and not Options.ModulesExtra ) then -- 3.3.5.1: Moved module options to options sub-tables
				Options.ModulesExtra = {};
			end

			private.Synchronize( Options ); -- Loads defaults if nil

		end
	end
end

Jughead:RegisterEvent( "ADDON_LOADED" );
